from collections import namedtuple

from .defaultpalette import default_palette
from .utils import chunks

Size = namedtuple('Size', 'x y z')
Color = namedtuple('Color', 'r g b a')
Voxel = namedtuple('Voxel', 'x y z c')
Model = namedtuple('Model', 'size voxels')
Material = namedtuple('Material', 'id type weight props')


def get_default_palette():
    return [Color(*tuple(i.to_bytes(4, 'little'))) for i in default_palette]


class Vox(object):

    def __init__(self, models, palette=None, materials=None):
        self.models = models
        self.default_palette = not palette
        self._palette = palette
        self.materials = materials or []

    @property
    def palette(self):
        return self._palette

    @palette.setter
    def palette(self, val):
        self._palette = val
        self.default_palette = False

    def to_dense_rgba(self, model_idx=0):

        import numpy as np
        m = self.models[model_idx]
        res = np.zeros((m.size.y, m.size.z, m.size.x, 4), dtype='B')

        for v in m.voxels:
            res[v.y, m.size.z - v.z - 1, v.x] = self.palette[v.c]

        return res

    def to_dense(self, model_idx=0):
        import numpy as np
        m = self.models[model_idx]
        res = np.zeros((m.size.x, m.size.y, m.size.z), dtype='B')

        for v in m.voxels:
            res[v.x, v.y, m.size.z - v.z - 1] = v.c

        return res

    def __str__(self):
        return f'Vox({self.models})'

    @staticmethod
    def from_dense(a, palette_thru, black=[0, 0, 0]):

        palette = None

        if len(a.shape) == 4:
            from PIL import Image
            import numpy as np

            mask = np.all(a == np.array([[black]]), axis=3)

            x, y, z, _ = a.shape

            # color index 0 is reserved for empty, so we get 255 colors
            img = Image.fromarray(a.reshape(x, y * z, 3)).quantize(255)
            palette = img.getpalette()
            palette = [Color(*c, 255) for c in chunks(palette, 3)]
            a = np.asarray(img, dtype='B').reshape(x, y, z).copy() + 1
            a[mask] = 0

        if len(a.shape) != 3:
            raise Exception("I expect a 4 or 3 dimensional matrix")

        y, z, x = a.shape

        nz = a.nonzero()

        voxels = [Voxel(nz[2][i], nz[0][i], z - nz[1][i] - 1, a[nz[0][i], nz[1][i], nz[2][i]]) for i in
                  range(nz[0].shape[0])]

        # return Vox([Model(Size(x, y, z), voxels)], palette)
        return Vox([Model(Size(x, y, z), voxels)], palette_thru)



from struct import pack
import logging
log = logging.getLogger()


class VoxWriter(object):

    def __init__(self, filename, vox):
        self.filename = filename
        self.vox = vox

    def _chunk(self, id, content, chunks=[]):

        res = b''
        for c in chunks:
            res += self._chunk(*c)

        return pack('4sii', id, len(content), len(res)) + content + res

    def _matflags(self, props):
        flags = 0
        res = b''
        for b, field in [(0, 'plastic'),
                         (1, 'roughness'),
                         (2, 'specular'),
                         (3, 'IOR'),
                         (4, 'attenuation'),
                         (5, 'power'),
                         (6, 'glow'),
                         (7, 'isTotalPower')]:
            if field in props:
                flags |= 1 << b
                try:
                    res += pack('f', props[field])
                except Exception as exc:
                    print(f"Bad prop content: {props[field]} ({type(props[field])}) expected float.")
                    raise exc

        return pack('i', flags) + res

    def write(self):

        res = pack('4si', b'VOX ', 150)

        chunks = []

        if len(self.vox.models) > 1:
            chunks.append((b'PACK', pack('i', len(self.vox.models))))

        for m in self.vox.models:
            chunks.append((b'SIZE', pack('iii', *m.size)))
            new_voxels = [pack('BBBB', *v) for v in m.voxels]
            new_chunk = pack('i', len(m.voxels)) + b''.join(new_voxels)
            log.debug(f"New chunk: {new_chunk}")
            chunks.append((b'XYZI', new_chunk))

        chunks.append((b'RGBA', b''.join(pack('BBBB', *c) for c in self.vox.palette)))

        for m in self.vox.materials:
            if m:
                chunks.append((b'MATT', pack('iif', m.id, m.type, m.weight) + self._matflags(m.props)))

        # TODO materials

        res += self._chunk(b'MAIN', b'', chunks)

        with open(self.filename, 'wb') as f:
            f.write(res)



import os
from unittest import TestCase

import numpy as np

from voxypy.models import Entity, validate_and_pad_palette, Color

HOWDY_FILE = os.path.sep.join(['test', 'howdy.vox'])
RED_FILE = os.path.sep.join(["test", "red.png"])
SINGLE_FILE = os.path.sep.join(["test", "single.vox"])
SINGLE_2_FILE = os.path.sep.join(["test", "single_2.vox"])
TINY_FILE = os.path.sep.join(['test', 'tiny.vox'])

def matrices_are_equal(m1, m2):
    return all([m1[x][y] == m2[x][y] for x in range(len(m1)) for y in range(len(m1[0]))])


def print_layers(a):
    s = ""
    for row in range(a.shape[2]):
        s += str(a[:, :, row])
        s += "\n\n"
    return s


def print_slices(a):
    s = ""
    for xy_slice in range(a.shape[1]):
        s += str(a[:, xy_slice, :])
        s += "\n\n"
    return s


class TestEntity(TestCase):

    def test_str(self):
        x, y, z = 1, 2, 3
        entity = Entity(x, y, z)
        entity.set(0, 0, 0, 255)
        self.assertEqual(str(entity), "Entity with dimensions x=1, y=2, z=3")

    def test_init_dense_bad_data(self):
        try:
            data = [[[0, 0, 0], [0, 0, 0]]]
            entity = Entity(3, 2, 1, data=data)
            self.fail("Should have failed with a mismatched data type.")
        except TypeError:
            pass
        except Exception as exc:
            self.fail(f"Unexpected exception type {exc}")

    def test_init_dense_bad_shape(self):
        shape = (1, 2, 3)
        arr = np.zeros(shape, dtype=int)
        try:
            entity = Entity(*shape[::-1], data=arr)
            self.fail("Should have failed with mismatched shape.")
        except ValueError:
            pass
        except Exception as exc:
            self.fail(f"Unexpected exception type {exc}")

    def test_init_dense_bad_dtype(self):
        shape = (1, 2, 3)
        try:
            data = np.zeros(shape, dtype=float)
            entity = Entity(data=data)
            self.fail("Should have failed with a mismatched data type.")
        except TypeError:
            pass
        except Exception as exc:
            self.fail(f"Unexpected exception type {exc}")

    def test_init_dense_success(self):
        shape = (1, 2, 3)
        arr = np.zeros(shape, dtype=int)
        entity = Entity(*shape, arr)
        expected = np.zeros(shape[:2], dtype=int)
        actual = entity.get_layer(0)
        self.assertTrue(np.array_equal(expected, actual), f"Expected: {expected}\n\nActual: {actual}")
        self.assertEqual((entity.x, entity.y, entity.z), shape)

    def test_get_all_voxels(self):
        x, y, z = 2, 2, 1
        entity = Entity(x, y, z)
        expected = [0, 0, 0, 0]
        actual = [vox for vox in entity.all_voxels()]
        self.assertEqual(expected, actual)

    def test_get_set(self):
        x, y, z = 1, 2, 3
        entity = Entity(x, y, z)
        entity.set(0, 1, 2, 255)
        self.assertEqual(entity.get(0, 1, 2), 255)

    def test_equality(self):
        x, y, z = 1, 2, 3
        entity = Entity(x, y, z)
        correct_entity = Entity(x, y, z)
        self.assertTrue(entity == correct_entity, f"{entity}\nshould equal\n{correct_entity}")
        incorrect_entity = Entity(x, y, z).set(0, 0, 0, 1)
        self.assertFalse(entity == incorrect_entity, f"{entity}\nshould not equal\n{incorrect_entity}")

    def test_get_layer(self):
        x, y, z = 1, 2, 3
        entity = Entity(x, y, z)
        expected = np.zeros((1, 2), dtype=int)
        actual = entity.get_layer(0)
        self.assertEqual(len(expected), x, "X dimension is incorrect.")
        self.assertEqual(len(expected[0]), y, "Y dimension is incorrect.")
        self.assertTrue(matrices_are_equal(expected, actual), "Expected %s\n\nbut received %s" % (expected, actual))
        for _x in range(x):
            for _y in range(y):
                entity.set(_x, _y, z=1, color=255)
                expected[_x][_y] = 255
        actual = entity.get_layer(1)
        self.assertTrue(matrices_are_equal(expected, actual), "Expected %s\n\nbut received %s" % (expected, actual))

    def test_set_layer(self):
        x, y, z = 1, 2, 3
        entity = Entity(x, y, z)
        layer = np.zeros((1, 2), dtype=int)
        layer[:][:] = 255
        entity.set_layer(z=0, layer=layer)
        actual = entity.get_layer(0)
        self.assertTrue(matrices_are_equal(layer, actual), "Expected %s\n\nbut received %s" % (layer, actual))

    def test_read_from_file(self):
        filename = HOWDY_FILE
        try:
            entity = Entity().from_file(filename)
        except Exception as exc:
            self.fail(f"Failed to read from file {filename} :: {exc}")

    def test_read_write(self):
        in_filename = SINGLE_FILE
        out_filename = SINGLE_2_FILE
        entity = Entity()
        entity2 = Entity()
        try:
            entity.from_file(in_filename)
        except Exception as exc:
            self.fail(f"Failed to read from file {in_filename} :: {exc}")

        try:
            print(f"Writing {out_filename}")
            entity.write(out_filename)
        except Exception as exc:
            self.fail(f"Failed to write file {out_filename} :: {exc}")

        try:
            entity2.from_file(out_filename)
        except Exception as exc:
            self.fail(f"Failed to read from file {out_filename} :: {exc}")

        self.assertEqual(entity, entity2, f"Entities should be the same before and after writing.\n"
                                          f"Entity 1:\n{entity}\n{print_slices(entity.get_dense())}\n"
                                          f"Entity 2:\n{entity2}\n{print_slices(entity2.get_dense())}")

        try:
            os.remove(out_filename)
        except Exception as exc:
            print(f"Failed to clean up after unit test. File: {out_filename}")

    def test_init_vs_file(self):
        tiny_mem = np.zeros((1, 2, 3), dtype=int)
        i = 1

        for z in range(3):
            for y in range(2):
                for x in range(1):
                    tiny_mem[x, y, z] = i
                    i += 1
        tiny_disk = Entity().from_file(TINY_FILE).get_dense()
        self.assertTrue(np.array_equal(tiny_mem, tiny_disk), f"\nMemory:\n{print_layers(tiny_mem)}"
                                                             f"\nDisk:\n{print_layers(tiny_disk)}")

    def test_read_palette_from_vox_file(self):
        entity = Entity().from_file(SINGLE_FILE)
        expected = [Color(75, 75, 75, 255)] * 255
        self.assertEqual(expected, entity.get_palette(padded=False))

    def test_read_palette_after_import(self):
        entity = Entity().from_file(SINGLE_FILE)
        entity.set_palette_from_file(RED_FILE)
        expected = [Color(255, 0, 0, 255)] * 255
        self.assertEqual(expected, entity.get_palette(padded=False))

    def test_validate_pad_palette_works(self):
        try:
            palette = [(i, i, i, i) for i in range(255)]
            validate_and_pad_palette(palette)
        except ValueError as exc:
            self.fail(f"Validation for working case failed: {exc}")

    def test_validate_pad_palette_pads(self):
        palette = [(i, i, i, i) for i in range(10)]
        validate_and_pad_palette(palette)
        self.assertEqual(255, len(palette), f"Function should return padded "
                                            f"palette with length 255 but received length {len(palette)}")

    def test_validate_pad_palette_too_long(self):
        try:
            palette = [(i, i, i, i) for i in range(257)]
            validate_and_pad_palette(palette)
            self.fail("Should have failed length check.")
        except ValueError:
            pass

    def test_validate_pad_palette_bad_data(self):
        try:
            palette = [(i, i, i, i) for i in range(255)]
            palette[0] = "invalid"
            validate_and_pad_palette(palette)
            self.fail("Should have failed bad data check.")
        except ValueError:
            pass

    def test_validate_pad_palette_bad_color(self):
        try:
            palette = [(i, i, i, i) for i in range(255)]
            palette[0] = (1, 2, 3)
            validate_and_pad_palette(palette)
            self.fail("Should have failed color length check.")
        except ValueError:
            pass

    def test_validate_pad_palette_bad_channel_data(self):
        try:
            palette = [(i, i, i, i) for i in range(255)]
            palette[0] = ("invalid", 2, 3, 255)
            validate_and_pad_palette(palette)
            self.fail("Should have failed channel data check.")
        except ValueError:
            pass

    def test_validate_pad_palette_bad_channel_low(self):
        try:
            palette = [(i, i, i, i) for i in range(255)]
            palette[0] = (-1, 2, 3, 255)
            validate_and_pad_palette(palette)
            self.fail("Should have failed channel data check.")
        except ValueError:
            pass

    def test_validate_pad_palette_bad_channel_high(self):
        try:
            palette = [(i, i, i, i) for i in range(255)]
            palette[0] = (1, 2, 256, 255)
            validate_and_pad_palette(palette)
            self.fail("Should have failed channel data check.")
        except ValueError:
            pass

    def test_nonzero(self):
        entity = Entity(data=np.ones(shape=(2, 2, 2), dtype=int))
        entity.set(0, 0, 0, 0)
        actual = entity.nonzero()
        expected = [(0, 0, 1), (0, 1, 1), (0, 1, 0), (1, 0, 0), (1, 0, 1), (1, 1, 1), (1, 1, 0)]
        print(actual)
        for i in range(len(expected)):
            self.assertTrue(actual[i] in expected)
        for i in range(len(actual)):
            self.assertTrue(expected[i] in actual)



from struct import unpack_from as unpack, calcsize, unpack_from
import logging

from .models import Vox, Size, Voxel, Model, Material

log = logging.getLogger(__name__)

MATERIALS = [(0, 'plastic'),
     (1, 'roughness'),
     (2, 'specular'),
     (3, 'IOR'),
     (4, 'attenuation'),
     (5, 'power'),
     (6, 'glow'),
     (7, 'isTotalPower')]

class ParsingException(Exception):
    pass


def bit(val, offset):
    mask = 1 << offset
    return val & mask


class Chunk(object):
    def __init__(self, chunk_id, content=None, chunks=None):
        self.id = chunk_id
        self.content = content or b''
        self.chunks = chunks or []
        self.material = None

        if chunk_id == b'MAIN':
            if len(self.content):
                raise ParsingException('Non-empty content for main chunk')
        elif chunk_id == b'PACK':
            self.models = unpack('i', content)[0]
        elif chunk_id == b'SIZE':
            self.size = Size(*unpack('iii', content))
        elif chunk_id == b'XYZI':
            # verified. issue must come when packing.
            log.debug(f"XYZI bytes: {content}")
            n = unpack('i', content)[0]
            log.debug(f'xyzi block with {n} voxels (len {len(content)})')
            self.voxels = []
            self.voxels = [Voxel(*unpack('BBBB', content, offset=4 + 4 * i)) for i in range(n)]
            log.debug(self.voxels)
        elif chunk_id == b'RGBA':
            self.palette = [(unpack('BBBB', content, 4 * i)) for i in range(255)]
            # Docs say:  color [0-254] are mapped to palette index [1-255]
            # hmm
            # self.palette = [ Color(0,0,0,0) ] + [ Color(*unpack('BBBB', content, 4*i)) for i in range(255) ]
        elif chunk_id == b'MATT' or chunk_id == b'MATL':
            _id, _type, weight, flags = unpack('iifi', content)
            props = {}
            offset = 16
            for b, field in MATERIALS:
                if bit(flags, b) and b < 7:  # no value for 7 / isTotalPower
                    props[field] = unpack('f', content, offset)[0]
                    offset += 4

            self.material = Material(_id, _type, weight, props)

        else:
            # raise ParsingException('Unknown chunk type: %s'%self.id)
            log.debug(f"Unknown chunk type {chunk_id}")
            pass

    def __str__(self):
        s = f"Chunk with id {self.id}"
        return s


class VoxParser(object):

    def __init__(self, filename):
        with open(filename, 'rb') as f:
            self.content = f.read()

        self.offset = 0

    def unpack(self, fmt):
        r = unpack(fmt, self.content, self.offset)
        self.offset += calcsize(fmt)
        return r

    def _parse_chunk(self):

        _id, n, m = self.unpack('4sii')

        log.debug("Found chunk id %s / len %s / children %s", _id, n, m)

        content = self.unpack('%ds' % n)[0]

        start = self.offset
        chunks = []
        while self.offset < start + m:
            chunks.append(self._parse_chunk())

        return Chunk(_id, content, chunks)

    def parse(self):

        header, version = self.unpack('4si')

        if header != b'VOX ':
            raise ParsingException("This doesn't look like a vox file to me")

        if version != 150:
            raise ParsingException("Unknown vox version: %s expected 150" % version)

        main = self._parse_chunk()

        if main.id != b'MAIN':
            raise ParsingException("Missing MAIN Chunk")

        chunks = list(reversed(main.chunks))
        if chunks[-1].id == b'PACK':
            models = chunks.pop().models
        else:
            models = 1

        log.debug("file has %d models", models)

        models = [self._parse_model(size=chunks.pop(), xyzi=chunks.pop()) for _ in range(models)]

        palette = None
        materials = []
        for chunk in chunks:
            if chunk.id == b'RGBA':
                log.debug("Parsed RGBA chunk")
                palette = chunk.palette
            elif chunk.id == b'MATT' or chunk.id == b'MATL':
                log.debug(f"Parsed {chunk.id} chunk")
                materials.append(chunk.material)
            else:
                log.debug(f"Skipping Chunk with id {chunk.id}")

        return Vox(models, palette, materials).to_dense(), palette

    def _parse_model(self, size, xyzi):
        if size.id != b'SIZE':
            raise ParsingException('Expected SIZE chunk, got %s', size.id)
        if xyzi.id != b'XYZI':
            raise ParsingException('Expected XYZI chunk, got %s', xyzi.id)

        return Model(size.size, xyzi.voxels)


if __name__ == '__main__':
    import sys
    import coloredlogs

    coloredlogs.install(level=logging.DEBUG)

    VoxParser(sys.argv[1]).parse()



import logging
from logging import StreamHandler, Formatter
import os
import sys
from unittest import TestCase

import numpy as np
from voxypy.numpy_vox_io.models import Vox

from voxypy.numpy_vox_io.parser import VoxParser
from voxypy.numpy_vox_io.writer import VoxWriter

TEST_DIR = os.path.sep.join([os.getcwd(), 'test'])
log = logging.getLogger()
log.setLevel(logging.DEBUG)
handler = StreamHandler(stream=sys.stdout)
handler.setFormatter(Formatter(fmt='[%(asctime)s: %(levelname)s\t] %(message)s'))
log.addHandler(handler)


class TestParser(TestCase):
    def test_read_from_file(self):
        filename = os.path.sep.join([TEST_DIR, 'howdy.vox'])
        try:
            voxels, palette = VoxParser(filename).parse()
        except Exception as exc:
            self.fail(f"Failed to read from file {filename} :: {exc}")

    def test_read_write(self):
        in_filename = os.path.sep.join([TEST_DIR, 'single.vox'])
        out_filename = os.path.sep.join([TEST_DIR, 'single2.vox'])
        vox = VoxParser(in_filename)
        try:
            log.info("Parsing...")
            vox, palette = vox.parse()
            vox = Vox.from_dense(vox, palette_thru=palette)
        except Exception as exc:
            self.fail(f"Failed to read from file {in_filename} :: {exc}")

        try:
            log.info("Writing...")
            VoxWriter(out_filename, vox).write()
        except Exception as exc:
            self.fail(f"Failed to write file {out_filename} :: {exc}")

        vox2 = None
        try:
            log.info("Parsing again...")
            vox2, palette = VoxParser(out_filename).parse()
            vox2 = Vox.from_dense(vox2, palette_thru=palette)
        except Exception as exc:
            self.fail(f"Failed to read from file {out_filename} :: {exc}")

        self.assertTrue(np.array_equal(vox.to_dense(), vox2.to_dense()),
                        "Entities should be the same before and after writing.")

        try:
            os.remove(out_filename)
        except Exception as exc:
            print(f"Failed to clean up after unit test. File: {out_filename}")
        # TODO: add comparisons for materials and such.



import logging
import os
import sys
from unittest import TestCase

TEST_DIR = os.path.sep.join([os.getcwd(), 'test'])
log = logging.getLogger()
log.setLevel(logging.DEBUG)
handler = logging.StreamHandler(stream=sys.stdout)
handler.setFormatter(logging.Formatter(fmt='[%(asctime)s: %(levelname)s\t] %(message)s'))
log.addHandler(handler)


class TestParser(TestCase):
    def test_read(self):
        # entity = Entity().
        pass


from unittest import TestCase

from voxypy.models import Voxel


class TestVoxel(TestCase):
    def test_set(self):
        v = Voxel(0)
        self.assertEqual(1, v.set(1))
        try:
            v.set(256)
            self.fail("Setting 256 should have failed.")
        except ValueError:
            pass
        except Exception as e:
            self.fail(f"Unexpected exception type {e}")
        try:
            v.set(-1)
            self.fail("Setting -1 should have failed.")
        except ValueError:
            pass
        except Exception as e:
            self.fail(f"Unexpected exception type {e}")

    def test_add(self):
        v = Voxel(0)
        a = int(v.add(1))
        self.assertEqual(0, a)
        v = Voxel(1)
        self.assertEqual(2, int(v.add(1)))
        v = Voxel(255)
        self.assertEqual(1, int(v.add(1)))
        v = Voxel(250)
        self.assertEqual(5, int(v.add(10)))
        v = Voxel(10)
        self.assertEqual(5, int(v.add(-5)))
        v = Voxel(1)
        self.assertEqual(255, int(v.add(-1)))
        v = Voxel(5)
        self.assertEqual(255, int(v.add(-5)))

    def test_subtract(self):
        v = Voxel(0)
        self.assertEqual(0, int(v.subtract(1)))
        v = Voxel(1)
        self.assertEqual(255, int(v.subtract(1)))



from collections import namedtuple

import imageio
import numpy as np

from voxypy.numpy_vox_io.models import Vox
from voxypy.numpy_vox_io.parser import VoxParser

import logging

from voxypy.numpy_vox_io.writer import VoxWriter

log = logging.getLogger()

Color = namedtuple('Color', ['r', 'g', 'b', 'a'])


def validate_and_pad_palette(palette):
    if Color(0, 0, 0, 0) in palette:
        palette.remove(Color(0, 0, 0, 0))
    if len(palette) > 255:
        raise ValueError(f"Invalid length of palette {len(palette)}. Expected 255 RGB 4-tuples.")
    elif len(palette) < 255:
        # pad incomplete palette
        palette.extend([(0, 0, 0, 0)] * (255 - len(palette)))
    for color in palette:
        if type(color) != tuple and type(color) != Color:
            raise ValueError(f"Failed to parse color {color}. Expected RGB 4-tuple or Color.")
        if len(color) != 4:
            raise ValueError(f"Failed to parse color {color}. Expected RGB 4-tuple or Color.")
        for channel in color:
            if type(channel) != int and type(channel) != np.int32 and type(channel) != np.int64:
                raise ValueError(f"Invalid type {type(channel)} in color {color}. Expected RGB 4-tuple of ints.")
            if channel < 0 or channel > 255:
                raise ValueError(f"Out of range value {channel} in color {color}. Expected RGB values to be 0-255.")
    return palette


class Voxel:
    def __init__(self, color=0):
        self._color = color

    def __eq__(self, other):
        if isinstance(other, Voxel):
            return self._color == other._color
        elif isinstance(other, int):
            return self._color == other

        try:
            return self._color == int(other)
        except Exception as e:
            raise ValueError(f"Unable to compare Voxel and {type(other)} :: {str(e)}")

    def __str__(self):
        return str(self._color)

    def __int__(self):
        return int(self._color)

    def subtract(self, amount=1):
        return self.add(-1 * amount)

    def add(self, amount=1):
        if self._color == 0:
            return self
        if not isinstance(amount, int):
            raise ValueError(f"Invalid increment type {type(amount)}")
        if self._color + amount < 0:
            amount += 1
        elif amount + self._color > 256:
            amount += 1
        self._color = (self._color + amount) % 256
        if self._color == 0:  # loop back 255 -> 1 or 1 -> 255 if subtracting
            if amount < 0:
                self._color = 255
            else:
                self._color = 1
        return self

    def set(self, color):
        color = int(color)
        if color < 0 or color > 255:
            raise ValueError(f"Invalid color index {color}. It must be 0-255")
        self._color = color
        return self

    def get(self):
        return self._color


class Entity:
    def __init__(self, x=0, y=0, z=0, data=None, palette=None):
        if data is None:
            self.x = x
            self.y = y
            self.z = z
            self._voxels = np.zeros((x, y, z), dtype=int)
        else:
            if type(data) != np.ndarray:
                raise TypeError(f"Expected `data` to be an NumPy ndarray, but received f{type(data)}")
            elif data.dtype != int:
                raise TypeError(f"Expected `data` have dtype=int, but received f{data.dtype}")
            elif data.shape != (x, y, z) and (x, y, z) != (0, 0, 0):
                raise ValueError(f"Expected `data` ndarray with shape ({x}, {y}, {z}) but found shape {data.shape}")
            self.from_dense(data)

        self._palette = None
        if palette:
            self.set_palette(palette)
        else:
            self.set_palette([Color(75, 75, 75, 255)] * 255)

    def __eq__(self, other):
        if not isinstance(other, Entity):
            return False
        if self.x != other.x or self.y != other.y or self.z != other.z:
            return False
        return np.array_equal(self.get_dense(), other.get_dense())

    def __str__(self):
        s = f"Entity with dimensions x={self.x}, y={self.y}, z={self.z}"
        return s

    def set_palette_from_file(self, filename):
        new_palette = np.array(imageio.imread(filename), dtype=int)
        new_palette = [Color(*c) for row in new_palette for c in row]
        self.set_palette(new_palette)

    def set_palette(self, palette):
        if palette is not None:
            palette = validate_and_pad_palette(palette)
        self._palette = palette

    def get_palette(self, padded=True):
        if padded:
            p = [(0, 0, 0, 0)]
            p.extend(self._palette)
            return p
        else:
            return self._palette

    def from_file(self, filename):
        voxels, palette = VoxParser(filename).parse()
        voxels = np.flip(voxels, axis=2)
        self._voxels = voxels
        self.set_palette(palette)
        self.x = len(self._voxels)
        self.y = len(self._voxels[0])
        self.z = len(self._voxels[0][0])
        return self

    def from_dense(self, dense):
        self.x = len(dense)
        self.y = len(dense[0])
        self.z = len(dense[0][0])
        self._voxels = dense
        return self

    def write(self, filename):
        # TODO: fix this on the back end
        writing_voxels = np.copy(self._voxels)
        writing_voxels = np.rot90(np.fliplr(np.swapaxes(writing_voxels, 0, 1)), k=3, axes=(1, 2))
        writing_voxels = np.flip(writing_voxels, axis=1)
        vox = Vox.from_dense(writing_voxels, palette_thru=self.get_palette(padded=False))
        VoxWriter(filename, vox).write()

    def save(self, filename):
        self.write(filename)

    def swap_axes(self, a, b):
        self._voxels = np.swapaxes(self._voxels, a, b)
        return self

    def rotate(self, k, axes):
        self._voxels = np.rot90(self._voxels, k=k, axes=axes)
        return self

    def get_dense(self):
        return self._voxels

    def all_voxels(self):
        return self._voxels.flatten()

    def flip(self):
        self._voxels = np.flip(self._voxels, 2)

    def get(self, x, y, z):
        if x >= self.x:
            raise IndexError(f"Dimension X out of bounds in get. Received {x} but max is {self.x - 1}")
        elif y >= self.y:
            raise IndexError(f"Dimension Y out of bounds in get. Received {y} but max is {self.y - 1}")
        elif z >= self.z:
            raise IndexError(f"Dimension Z out of bounds in get. Received {z} but max is {self.z - 1}")

        return Voxel(self._voxels[x, y, z])

    def set(self, x, y, z, color):
        color = int(color)
        if color < 0 or color > 255:
            raise IndexError(f"Color {color} out of bounds [0-255]")
        self._voxels[x, y, z] = color

    def get_layer(self, z):
        if z > self.z:
            raise IndexError(f"Dimension Z out of bounds in get_layer. Received {z} but max is {self.z}")
        return self._voxels[:, :, z]

    def set_layer(self, z, layer):
        if type(layer) != np.ndarray:
            raise TypeError(f"Invalid input type {type(layer)}, please use np.ndarray")
        elif layer.shape != (self.x, self.y):
            raise IndexError(f"Mismatched dimensions in set_layer."
                             f" Received ({len(layer)}, {len(layer[0])})"
                             f" but expected ({self.x}, {self.y})")
        elif self._voxels.dtype != layer.dtype:
            raise TypeError(f"Layer has dtype of {layer.dtype}, but should be {self._voxels.dtype}")
        log.debug(f"Setting layer {z} to {layer}")
        self._voxels[:, :, z] = layer

    def layers(self):
        for _z in range(self.z):
            yield self.get_layer(_z)

    def nonzero(self):
        hits = np.where(self._voxels != 0)
        hits = list(zip(*hits))
        return hits



import pathlib
from setuptools import setup, find_packages

# The directory containing this file
HERE = pathlib.Path(__file__).parent

# The text of the README file
README = (HERE / "README.md").read_text()

# This call to setup() does all the work
setup(
    name="voxypy",
    version="0.2.3",
    description="Data structures for Voxel operations.",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://gitlab.com/Claytone/voxypy",
    author="Claytone",
    author_email="clayton.wells.321@gmail.com",
    license="MIT",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
    ],
    packages=find_packages(exclude=("test",)),
    include_package_data=True,
    install_requires=["numpy", "Pillow", "imageio"],
    entry_points={},
)